import os
from typing import Dict

import numpy as np
import pydicom

KEY_DIM_1 = "!matrix size [1]"
KEY_DIM_2 = "!matrix size [2]"
KEY_DIM_3 = "!matrix size [3]"
KEY_NPROJECTIONS = "!number of projections"

KEY_DATA_FILE = "name of data file"

KEY_BYTES_PER_PIXEL = "!number of bytes per pixel"
KEY_NUMBER_FORMAT = "!number format"


def type_by_format(number_format: str, bytes_per_pixel: int) -> type:
    if number_format == "float" and bytes_per_pixel == 4:
        return np.float32

    raise ValueError("Unknown mapping from format {number_format} with {bytes_per_pixel} bytes to numpy type")


def parse_interfile_header(filename: str) -> Dict[str, str]:
    with open(filename, mode="r") as f:
        lines = f.readlines()

    items = map(lambda line: line.split(":="), lines)
    items = filter(lambda item: len(item) == 2, items)
    items = map(lambda item: [item[0].strip(), item[1].strip()], items)
    return dict(items)


def read_interfile(filename: str) -> np.ndarray:
    header = parse_interfile_header(filename)

    dim_x = int(header[KEY_DIM_1])
    dim_y = int(header[KEY_DIM_2])
    dim_z = int(header[KEY_DIM_3]) if KEY_DIM_3 in header else int(header[KEY_NPROJECTIONS])

    bytes_per_pixel = int(header[KEY_BYTES_PER_PIXEL])
    num_format = header[KEY_NUMBER_FORMAT]
    dtype = type_by_format(num_format, bytes_per_pixel)

    data_file = os.path.join(os.path.dirname(filename), header[KEY_DATA_FILE])
    with open(data_file, mode="rb") as f:
        image = np.frombuffer(f.read(), dtype)
    image = image.reshape((dim_z, dim_y, dim_x))
    return image, header
    

UINT16_MAX = 2**16 - 1
def scale_image(image: np.ndarray, initial_scale=10000000):
    scale = initial_scale
    while (scale * image.max()) > UINT16_MAX:
        scale = scale / 10
    image = (image * scale).astype(np.uint16)
    return image, scale


def update_dcm(dcm: pydicom.dataset.FileDataset, image: np.ndarray) -> pydicom.dataset.FileDataset:
    image, scale = scale_image(image)

    dcm.NumberOfFrames = image.shape[0] 
    dcm.NumberOfSlices = image.shape[0] 
    dcm.SliceVector = list(range(1, image.shape[0] + 1))
    dcm.Columns = image.shape[1]
    dcm.Rows = image.shape[2]
    dcm.PixelData = image.tobytes()
    dcm.WindowWidth = image.max()
    dcm.WindowCenter = image.max() / 2
    dcm.LargestImagePixelValue = image.max()
    dcm[0x00331038].value = scale
    return dcm


if __name__ == "__main__":
    import argparse

    import pydicom

    parser = argparse.ArgumentParser(description="convert an INTERFILE image to a DICOM image", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--inter", type=str, required=True, help="the INTERFILE image to be converted")
    parser.add_argument("--dcm", type=str, required=True, help="a DICOM reference image - data and according fields are exchanged from the INTERFILE image")
    parser.add_argument("--out", type=str, help="the output file - defaults to the INTERFILE input with a dcm extension")
    args = parser.parse_args()

    if args.out is None:
        args.out = os.path.splitext(args.inter)[0] + ".dcm"

    image, header = read_interfile(args.inter)
    image = image[47:80, :, ::-1]

    dcm = pydicom.dcmread(args.dcm)
    _dcm = update_dcm(dcm.copy(), image)

    pydicom.dcmwrite(args.out, _dcm)

